/*****************************************************************
au record & play (duplex, for Linux)

$Id: auduplex.c,v 1.4 2000/06/06 23:29:22 purnhage Exp $

Source file: auduplex.c

(c) Heiko Purnhagen 1998, 1999, 2000
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.tnt.uni-hannover.de/~purnhage/

All usual disclaimers apply!

This program may be copied according to the GNU General Public Licence.
See file COPYING for further details.

Authors:
HP    Heiko Purnhagen, Uni Hannover <purnhage@tnt.uni-hannover.de>

Changes:
31-mai-00   HP      first version, based on plax and auecho
06-jun-00   HP      added non-blocking au i/o
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <ctype.h>

#include "common.h"		/* common module */
#include "cmdline.h"		/* command line module */
#include "austream.h"		/* audio i/o streams */

#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/soundcard.h>


/* ---------- declarations ---------- */

#define PROGVER "auduplex V0.9.0 06-Jun-2000"
#define CVSID "$Id: auduplex.c,v 1.4 2000/06/06 23:29:22 purnhage Exp $"
#ifndef DISTVER		/* set in Makefile */
#define DISTVER "<unknown>"
#endif

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define STRLEN (255)

#define FS_MAX (48000.0)
#define BUF_TIME (1.)		/* length of buffer in seconds */
#define MAX_CH (16)

#ifndef PI
#define PI 3.14159265358979323846264338327950288
#endif


/* ---------- variables ---------- */

/* command line module */

static char *progName = "<no program name>";
static char *recFileName;
static int numChan;
static float fSample;
static char *audioDevice;
static float gainRec;
static float gainPlay;
static float delayTime;
static float duration;
static float bufTime;
static int recNB;
static int playNB;

static int debugLevel;
static int *varArgIdx;

static CmdLinePara paraList[] = {
  {&varArgIdx,NULL,"<play input .au file (- = stdin)>"},
  {NULL,NULL,NULL}
};

static CmdLineSwitch switchList[] = {
  {"h",NULL,NULL,NULL,NULL,"print help"},
  {"o",&recFileName,"%s","-",NULL,
   "record output .au file (- = stdout)"},
  {"c",&numChan,"%d","1",NULL,
   "number of channels"},
  {"s",&fSample,"%f","8000.0",NULL,
   "sampling frequency [Hz]"},
  {"ad",&audioDevice,"%s","/dev/dsp",NULL,
   "audio device name"},
  {"gr",&gainRec,"%f","1.0",NULL,
   "rec. amplitude gain"},
  {"gp",&gainPlay,"%f","1.0",NULL,
   "play amplitude gain"},
  {"ts",&delayTime,"%f","1.0",NULL,
   "start time (delay before play) [sec]"},
  {"td",&duration,"%f","0.0",NULL,
   "max. duration [sec] (0=unlimited)"},
  {"b",&bufTime,"%f","0.2",NULL,
   "buffer size [sec]"},
  {"nr",&recNB,NULL,NULL,NULL,
   "record with non-blocking output file"},
  {"np",&playNB,NULL,NULL,NULL,
   "play with non-blocking input file"},
  {"d",&debugLevel,"%i","0",NULL,
   "debug level (on stderr) (0=off 1=basic 2=med 3=high)\n"
   "(+4*austreamDbgLvl)"},
  {NULL,NULL,NULL,NULL,NULL,NULL}
};

static int stop = 0;


/* ---------- local functions ---------- */

void stopproc (int sig)
{
  stop = 1;
}


/* ---------- main ---------- */

int main (int argc, char *argv[])
{
  char playFileName[STRLEN];
  AuStream *inas,*outas;
  int audev;
  int result;

  int inNumChan;
  float inFSample;
  long inNumSamp;
  long maxBufLen;
  long bufLen;
  long bufTrans;
  long numClip;
  long recClip;
  long playClip;
  short *bufR;
  short *bufP;
  long i,tmp;
  int recPeak,playPeak;
  char level[17];

  long recNumSamp;
  long playNumSamp;
  long delayNumSamp;
  long durNumSamp;
  int samplerate,channels;
  int caps,err,triggered;

  /* evaluate command line  */
  CmdLineInit(0);
  result = CmdLineEval(argc,argv,paraList,switchList,1,&progName);
  if (result) {
    if (result==1) {
      fprintf(stderr,"%s: %s\n"
	      "%s\n",
	      progName,PROGVER "   " DISTVER,CVSID);
      CmdLineHelp(progName,paraList,switchList,stderr);
      fprintf(stderr,
	      "example (tcsh):\n"
	      "  mkfifo fifo.au\n"
	      "  cat fifo.au | linux6/fifo | linux6/auduplex -d 2 - | "
	      "cat >! fifo.au\n");
      exit (1);
    }
    else
      CommonExit(1,"command line error (\"-h\" for help)");
  }
  AuInit((debugLevel>>2)&3);
  debugLevel &= 3;

  if (debugLevel)
    fprintf(stderr,"%s: %s\n",progName,PROGVER "   " DISTVER);
  if (debugLevel>1) {
    fprintf(stderr,"%s\n",CVSID);
    fprintf(stderr,"debugLevel=%d\n",debugLevel);
  }

  if (varArgIdx[0]>=0) {
    if (varArgIdx[1]>=0)
      CommonExit(1,"max one input file allowed");
      strncpy(playFileName,argv[varArgIdx[0]],STRLEN);
  }
  else
    /* strcpy(playFileName,"-"); */
    CommonExit(1,"input file required");
  
  if (delayTime < 0)
    delayTime = 0;
  delayNumSamp = numChan*(long)(fSample*delayTime+.5);
  if (duration < 0)
    duration = 0;
  durNumSamp = numChan*(long)(fSample*duration+.5);

  maxBufLen = (long)(BUF_TIME*FS_MAX*2);
  if(!(bufR = (short *)malloc(maxBufLen*sizeof(short))))
    CommonExit(1,"failed to allocate bufR");
  if(!(bufP = (short *)malloc(maxBufLen*sizeof(short))))
    CommonExit(1,"failed to allocate bufP");

  bufLen = (long)(numChan*fSample*bufTime+.5);
  bufLen = max(bufLen,numChan);
  bufLen = (bufLen/numChan)*numChan;

  /* signal(SIGINT,stopproc); */

  if (debugLevel)
    fprintf(stderr,
	    "%s: audio device -> %s & %s -> audio device\n",progName,
	    recFileName,playFileName);

  /* open output file */
  if (!(outas=AuOpenWriteNB(recFileName,recNB,numChan,fSample)))
    CommonExit(1,"can not open output \"%s\"",recFileName);

  if (debugLevel>1)
    fprintf(stderr,
	    "output=%s numChan=%d fSample=%.1f\n",
	    recFileName,numChan,fSample);

  /* open audio device */
  samplerate = (int)(fSample+.5);
  channels = numChan;
  if ((audev=open(audioDevice,O_RDWR,0))==-1)
    CommonExit(1,"Can not open \"%s\"\n",audioDevice);
  ioctl(audev, SNDCTL_DSP_SETDUPLEX, 0);
  ioctl(audev, SNDCTL_DSP_GETCAPS, &caps);
  if (debugLevel>1)
    fprintf(stderr,"caps=%d dup=%d (%d) trig=%d (%d)\n",
	    caps,DSP_CAP_DUPLEX,(caps&DSP_CAP_DUPLEX)?1:0,
	    DSP_CAP_TRIGGER,(caps&DSP_CAP_TRIGGER)?1:0);
  tmp = ~PCM_ENABLE_OUTPUT & ~PCM_ENABLE_INPUT; /* disable */
  ioctl(audev, SNDCTL_DSP_SETTRIGGER, &tmp);

  /* use 20 ms .. 40 ms fragement size */
  tmp = 9;				/* min 512 bytes/fragment */
  while (1<<tmp < (int)(0.020*samplerate+.5)*2*(channels))
    tmp++;
  tmp |= 0x7fff0000;
  err = ioctl(audev, SNDCTL_DSP_SETFRAGMENT, &tmp);

  tmp = AFMT_S16_LE;			/* aka 16 */
  err = ioctl(audev, SNDCTL_DSP_SAMPLESIZE, &tmp);
  if (tmp != AFMT_S16_LE)
    CommonExit(1,"16 bit audio not supported\n");
  tmp = (channels>1)?1:0;
  err = err || ioctl(audev, SNDCTL_DSP_STEREO, &tmp);
  channels = tmp+1;
  tmp = samplerate;
  err = err || ioctl (audev, SNDCTL_DSP_SPEED, &tmp);
  samplerate = tmp;
  if (err)
    CommonExit(1," Unable to init dsp\n");
  if (debugLevel>1)
    fprintf (stderr,"audio device: samplerate=%d channels=%d\n",
	     samplerate,channels);
  if (channels!=numChan && samplerate!=fSample)
    CommonExit(1,"wrong input file format (numChan / fSample)");

  recNumSamp = 0;
  recClip = 0;
  playNumSamp = 0;
  playClip = 0;
  triggered = 0;
  for (i=0; i<bufLen; i++)
    bufP[i] = 0;

  /* record delayTime samples */
  do {
    if (debugLevel) {
      if (debugLevel<=2) {
	fprintf(stderr,"\r%8.3f %8.3f ",recNumSamp/numChan/fSample,
		playNumSamp/numChan/fSample);
	fflush(stderr);
      }
      else
	fprintf(stderr,"%8.3f %8.3f ",recNumSamp/numChan/fSample,
		playNumSamp/numChan/fSample);
    }

    bufTrans = delayNumSamp-recNumSamp;
    if (bufTrans>bufLen)
      bufTrans = bufLen;

    write(audev,bufP,bufTrans*sizeof(short));
    if (!triggered) {
      tmp = PCM_ENABLE_OUTPUT | PCM_ENABLE_INPUT; /* enable */
      ioctl(audev, SNDCTL_DSP_SETTRIGGER, &tmp);
      triggered = 1;
    }
    if (debugLevel>=2) { fprintf(stderr,"R"); fflush(stderr); }
    tmp = read(audev,bufR,bufTrans*sizeof(short));
    if (debugLevel>=2) { fprintf(stderr,"+"); fflush(stderr); }
    if (tmp!=(long)(bufTrans*sizeof(short)))
      CommonExit(1,"got only %d/%d bytes\n",tmp,bufTrans*sizeof(short));

    numClip = recPeak = 0;
    for (i=0; i<bufTrans; i++) {
      tmp = ((long)(gainRec*bufR[i]+32768.5))-32768;
      if (tmp>32767) {
	tmp = 32767;
	numClip++;
      }
      if (tmp<-32768) {
	tmp = -32768;
	numClip++;
      }
      if (recPeak < tmp)
	recPeak = tmp;
      if (recPeak < -tmp)
	recPeak = -tmp;
      bufR[i] = tmp;
    }
    recClip += numClip;

    if (debugLevel>=2) { fprintf(stderr,"o"); fflush(stderr); }
    AuWriteData(outas,bufR,bufTrans);
    if (debugLevel>=2) { fprintf(stderr,"+ "); fflush(stderr); }

    if (debugLevel) {
      for (i=0; i<16; i++) {
	level[i] = (recPeak)?'*':'-';
	recPeak >>= 1;
      }
      level[16]='\0';
      fprintf(stderr,"%s ---------------- ",level);
      if (debugLevel<=2)
	fflush(stderr);
      else
	fprintf(stderr," \n");
    }

    recNumSamp += bufTrans;
  } while (recNumSamp < delayNumSamp);

  /* open input file */
  if (!(inas=AuOpenReadNB(playFileName,playNB,
			  &inNumChan,&inFSample,&inNumSamp)))
    CommonExit(1,"can not open input \"%s\"",playFileName);

  if (debugLevel>1)
    fprintf(stderr," \n"
	    "input=%s inNumChan=%d inFSample=%.1f inNumSamp=%ld\n",
	    recFileName,inNumChan,inFSample,inNumSamp);

  if (inNumChan!=numChan && inFSample!=fSample)
    CommonExit(1,"wrong input file format (numChan / FSample)");

  /* echo max. duration samples */
  do {
    if (debugLevel) {
      if (debugLevel<=2) {
	fprintf(stderr,"\r%8.3f %8.3f ",recNumSamp/numChan/fSample,
		playNumSamp/numChan/fSample);
	fflush(stderr);
      }
      else
	fprintf(stderr,"%8.3f %8.3f ",recNumSamp/numChan/fSample,
		playNumSamp/numChan/fSample);
    }

    if (durNumSamp!=0)
      bufTrans = durNumSamp-playNumSamp;
    else
      bufTrans = bufLen;
    if (bufTrans>bufLen)
      bufTrans = bufLen;

    if (debugLevel>=2) { fprintf(stderr,"i"); fflush(stderr); }
    tmp = AuReadData(inas,bufP,bufTrans);
    if (debugLevel>=2) { fprintf(stderr,"+"); fflush(stderr); }
    if (tmp!=bufTrans)
      CommonExit(1,"got only %d/%d shorts from input file\n",
		 tmp,bufTrans);

    numClip = playPeak = 0;
    for (i=0; i<bufTrans; i++) {
      tmp = ((long)(gainPlay*bufP[i]+32768.5))-32768;
      if (tmp>32767) {
	tmp = 32767;
	numClip++;
      }
      if (tmp<-32768) {
	tmp = -32768;
	numClip++;
      }
      if (playPeak < tmp)
	playPeak = tmp;
      if (playPeak < -tmp)
	playPeak = -tmp;
      bufP[i] = tmp;
    }
    playClip += numClip;

    if (debugLevel>=2) { fprintf(stderr,"P"); fflush(stderr); }
    write(audev,bufP,bufTrans*sizeof(short));
    if (debugLevel>=2) { fprintf(stderr,"+"); fflush(stderr); }
    if (!triggered) {
      tmp = PCM_ENABLE_OUTPUT | PCM_ENABLE_INPUT; /* enable */
      ioctl(audev, SNDCTL_DSP_SETTRIGGER, &tmp);
      triggered = 1;
    }
    if (debugLevel>=2) { fprintf(stderr,"R"); fflush(stderr); }
    tmp = read(audev,bufR,bufTrans*sizeof(short));
    if (debugLevel>=2) { fprintf(stderr,"+"); fflush(stderr); }
    if (tmp!=(long)(bufTrans*sizeof(short)))
      CommonExit(1,"got only %d/%d bytes from audio device\n",
		 tmp,bufTrans*sizeof(short));

    numClip = recPeak = 0;
    for (i=0; i<bufTrans; i++) {
      tmp = ((long)(gainRec*bufR[i]+32768.5))-32768;
      if (tmp>32767) {
	tmp = 32767;
	numClip++;
      }
      if (tmp<-32768) {
	tmp = -32768;
	numClip++;
      }
      if (recPeak < tmp)
	recPeak = tmp;
      if (recPeak < -tmp)
	recPeak = -tmp;
      bufR[i] = tmp;
    }
    recClip += numClip;

    if (debugLevel>=2) { fprintf(stderr,"o"); fflush(stderr); }
    AuWriteData(outas,bufR,bufTrans);
    if (debugLevel>=2) { fprintf(stderr,"+ "); fflush(stderr); }

    playNumSamp += bufTrans;
    recNumSamp += bufTrans;

    if (debugLevel) {
      for (i=0; i<16; i++) {
	level[i] = (recPeak)?'*':'-';
	recPeak >>= 1;
      }
      level[16]='\0';
      fprintf(stderr,"%s ",level);
      for (i=0; i<16; i++) {
	level[i] = (playPeak)?'*':'-';
	playPeak >>= 1;
      }
      level[16]='\0';
      fprintf(stderr,"%s ",level);
      if (debugLevel<=2)
	fflush(stderr);
      else
	fprintf(stderr," \n");
    }
  } while (!stop && (durNumSamp==0 || playNumSamp < durNumSamp));

  /* close audio device */ 
  ioctl(audev, SNDCTL_DSP_SYNC, 0);

  /* close files */
  AuClose(outas);
  AuClose(inas);
  
  if (debugLevel && debugLevel<=2) {
    fprintf(stderr," \n");
    fflush(stderr);
  }
  if (debugLevel>1)
    fprintf(stderr,
	    "rec: %ld samples (%.3f sec)   "
	    "play: %ld samples (%.3f sec)\n",
	    recNumSamp,recNumSamp/(numChan*fSample),
	    playNumSamp,playNumSamp/(numChan*fSample));
  if (recClip)
    CommonWarning("rec: %ld samples clipped!",recClip);
  if (playClip)
    CommonWarning("play: %ld samples clipped!",playClip);

  CmdLineEvalFree(paraList);

  if (debugLevel)
    fprintf(stderr,"%s: %s\n",progName,stop?"aborted":"finished");

  exit (0);
}

/* end of auduplex.c */
