/**********************************************************************
audio i/o streams (.au format)

$Id: austream.c,v 1.8 2000/09/11 00:42:10 purnhage Exp $

Source file: austream.c

Authors:
HP    Heiko Purnhagen, Uni Hannover <purnhage@tnt.uni-hannover.de>
NM    Nikolaus Meine, Uni Hannover <meine@tnt.uni-hannover.de>

Changes:
16-sep-98   HP/NM   born, based on au_io.c
06-aug-99   HP      fixed bug in AuOpenRead() numSample
06-jun-00   HP      added non-blocking i/o: AuOpenReadNB(), AuOpenWriteNB()
04-sep-00   HP      cleaned up
10-sep-00   HP      fixed AuReadData() EOF-handling for stdin
**********************************************************************/

/* Audio i/o streaming with stdin/stdout support. */
/* Header format: .au (Sun audio file format, aka SND or AFsp) */
/* Sample format: 16 bit twos complement, uniform quantisation */
/* Data size set to -1 (=unknown) */

/* Multi channel data is interleaved: l0 r0 l1 r1 ... */
/* Total number of samples (over all channels) is used. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "austream.h"
#include "common.h"


/* ---------- declarations ---------- */

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))


/* ---------- declarations (structures) ---------- */

struct AuStreamStruct		/* audio stream handle */
{
  FILE *f;			/* stream handle */
  int fd;			/* -1 or file descriptor if non-blocking */
  long currentSample;		/* number of samples read/written */
  int eof;			/* eof/error flag */
  int write;			/* write (not read) flag */
};


/* ---------- variables ---------- */

static int AUdebugLevel = 0;	/* debug level */


/* ---------- local functions ---------- */

static void putshortB (short x, AuStream *s)
{
  register int a;

  putc((x>>8)&255,s->f);
  a = putc(x&255,s->f);
  if (a==EOF)
    s->eof = 1;
}

static short getshortB (AuStream *s)
{
  register int a,b;
  
  if (s->eof)
    return 0;
  
  a = getc(s->f);
  b = getc(s->f);
  if (b==EOF) {
    s->eof = 1;
    return 0;
  }
  return (a<<8)|b;
}

static int putcB (int c, AuStream *s)
{
  unsigned char a;

  if (s->fd==-1)
    return(putc(c,s->f));
  else {
    a = c&255;
    return((write(s->fd,&a,1)!=1)?EOF:a);
  }
}

static int getcB (AuStream *s)
{
  unsigned char a;

  if (s->fd==-1)
    return(getc(s->f));
  else
    return((read(s->fd,&a,1)!=1)?EOF:a);
}

#ifdef OBSOLETE
static void putshort (short x, AuStream *s)
{
  register int a;

  putcB((x>>8)&255,s);
  a = putcB(x&255,s);
  if (a==EOF)
    s->eof = 1;
}
#endif

static void putint (long x, AuStream *s)
{
  int a;

  putcB((x>>24)&255,s);
  putcB((x>>16)&255,s);
  putcB((x>>8)&255,s);
  a = putcB(x&255,s);
  if (a==EOF)
    s->eof = 1;
}

#ifdef OBSOLETE
static short getshort (AuStream *s)
{
  register int a,b;
  
  if (s->eof)
    return 0;
  
  a = getcB(s);
  b = getcB(s);
  if (b==EOF) {
    s->eof = 1;
    return 0;
  }
  return (a<<8)|b;
}
#endif

static long getint (AuStream *s)
{
  int a,b;

  if (s->eof)
    return 0;
  
  a = getcB(s)<<24;
  a |= getcB(s)<<16;
  a |= getcB(s)<<8;
  b = getcB(s);
  if (b==EOF) {
    s->eof = 1;
    return 0;
  }
  return a|b;
}

static char getbyte (AuStream *s)
{
  int a;

  if (s->eof)
    return 0;
  
  a = getcB(s);
  if (a==EOF) {
    s->eof = 1;
    return 0;
  }
  return a;
}


/* ---------- functions ---------- */

/* AuInit() */
/* Init audio i/o streams. */

void AuInit (
  int debugLevel)		/* in: debug level */
				/*     0=off  1=basic  2=full */
{
  AUdebugLevel = debugLevel;
  if (AUdebugLevel) {
    printf("AuInit: debugLevel=%d\n",AUdebugLevel);
  }
}


/* AuOpenRead() */
/* Open audio stream for reading. */

AuStream *AuOpenRead (
  char *streamName,		/* in: stream name, "-" for stdin */
  int *numChannel,		/* out: number of channels */
  float *fSample,		/* out: sampling frequency [Hz] */
  long *numSample)		/* out: number of samples in stream */
				/*      or -1 if not available */
				/* returns: */
				/*  audio stream (handle) */
				/*  or NULL if error */
{
  return(AuOpenReadNB(streamName,0,numChannel,fSample,numSample));
}


/* AuOpenWrite() */
/* Open audio stream for writing. */

AuStream *AuOpenWrite (
  char *streamName,		/* in: stream name, "-" for stdout */
  int numChannel,		/* in: number of channels */
  float fSample)		/* in: sampling frequency [Hz] */
				/* returns: */
				/*  audio stream (handle) */
				/*  or NULL if error */
{
  return(AuOpenWriteNB(streamName,0,numChannel,fSample));
}


/* AuOpenReadNB() */
/* Open audio stream for reading, non-blocking mode possible. */

AuStream *AuOpenReadNB (
  char *streamName,		/* in: stream name, "-" for stdin */
  int nonBlocking,		/* in: 1 for non-blocking mode */
  int *numChannel,		/* out: number of channels */
  float *fSample,		/* out: sampling frequency [Hz] */
  long *numSample)		/* out: number of samples in stream */
				/*      or -1 if not available */
				/* returns: */
				/*  audio stream (handle) */
				/*  or NULL if error */
{
  AuStream *s;
  long h,ofs,dsize,formc,srate,nchan;

  if (AUdebugLevel)
    printf("AuOpenReadNB: fileName=\"%s\"\n",streamName);

  if ((s=(AuStream*)malloc(sizeof(AuStream)))==NULL)
    CommonExit(-1,"AuOpenReadNB: Can not allocate memory");
  s->currentSample = 0;
  s->eof = 0;
  s->write = 0;

  if (nonBlocking) {
    if (streamName[0]=='-' && streamName[1]==0)
      s->fd = 0;   /* stdin */
    else
      s->fd = open(streamName,O_RDONLY);
    if (s->fd==-1) {
      CommonWarning("AuOpenReadNB: Can not open \"%s\"",streamName);
      free(s);
      return NULL;
    }
  }
  else {
    s->fd = -1;
    if (streamName[0]=='-' && streamName[1]==0)
      s->f = stdin;
    else
      s->f = fopen(streamName,"rb");
    if (s->f==NULL) {
      CommonWarning("AuOpenReadNB: Can not open \"%s\"",streamName);
      free(s);
      return NULL;
    }
  }

  h = getint(s);
  if (h!=0x2e736e64) {		/* magic string: .snd */
    CommonWarning("AuOpenReadNB: Wrong magic string in \"%s\"",streamName);
    AuClose(s);
    return NULL;
  }
  ofs = getint(s);
  dsize = getint(s);
  formc = getint(s);		/* 3 = 16 bit uniform 2compl */
  srate = getint(s);
  nchan = getint(s);
  for (h=24; h<ofs; h++)
    getbyte(s);
  if (s->eof || nchan<1 || formc!=3) {
    CommonWarning("AuOpenReadNB: Unsupported audio format in \"%s\"",streamName);
    AuClose(s);
    return NULL;
  }

  *numChannel = nchan;
  *fSample = srate;
  *numSample = (dsize<0)?-1:dsize/2;

  if (AUdebugLevel)
    printf("AuOpenReadNB: numChannel=%d  fSample=%.1f numSample=%ld\n",
	   *numChannel,*fSample,*numSample);

  if (s->fd!=-1)
    fcntl(s->fd,F_SETFL,fcntl(s->fd,F_GETFL,0)|O_NONBLOCK);

  return s;
}


/* AuOpenWriteNB() */
/* Open audio stream for writing, non-blocking mode possible. */

AuStream *AuOpenWriteNB (
  char *streamName,		/* in: stream name, "-" for stdout */
  int nonBlocking,		/* in: 1 for non-blocking mode */
  int numChannel,		/* in: number of channels */
  float fSample)		/* in: sampling frequency [Hz] */
				/* returns: */
				/*  audio stream (handle) */
				/*  or NULL if error */
{
  AuStream *s;

  if (AUdebugLevel) {
    printf("AuOpenWriteNB: fileName=\"%s\"\n",streamName);
    printf("AuOpenWriteNB: numChannel=%d  fSample=%.1f\n",
	   numChannel,fSample);
  }

  if ((s=(AuStream*)malloc(sizeof(AuStream)))==NULL)
    CommonExit(-1,"AuOpenWriteNB: Can not allocate memory");
  s->currentSample = 0;
  s->eof = 0;
  s->write = 1;

  if (nonBlocking) {
    if (streamName[0]=='-' && streamName[1]==0)
      s->fd = 1;   /* stdout */
    else
      s->fd = open(streamName,O_WRONLY|O_CREAT);
    if (s->fd==-1) {
      CommonWarning("AuOpenWriteNB: Can not open \"%s\"",streamName);
      free(s);
      return NULL;
    }
  }
  else {
    s->fd = -1;
    if (streamName[0]=='-' && streamName[1]==0)
      s->f = stdout;
    else
      s->f = fopen(streamName,"wb");
    if (s->f==NULL) {
      CommonWarning("AuOpenWriteNB: Can not open \"%s\"",streamName);
      free(s);
      return NULL;
    }
  }

  putint(0x2e736e64,s);	/* magic string: .snd */
  putint(28,s);		/* header size */
  putint(-1,s);		/* -1 = data size unknown */
  putint(3,s);		/* 3 = 16 bit uniform 2compl */
  putint((int)(fSample+.5),s);
  putint(numChannel,s);
  putint(0,s);		/* info string	*/

  if (s->eof) {
    CommonWarning("AuOpenWriteNB: Can not write to \"%s\"",streamName);
    AuClose(s);
    return NULL;
  }

  if (s->fd!=-1)
    fcntl(s->fd,F_SETFL,fcntl(s->fd,F_GETFL,0)|O_NONBLOCK);

  return s;
}


/* AuReadData() */
/* Read data from audio stream. */
/* if non-blocking: fill with 0 if samples can not be read */

long AuReadData (
  AuStream *stream,		/* in: audio stream (handle) */
  short *data,			/* out: data[] */
  long numSample)		/* in: number of samples to be read */
				/* returns: */
				/*  number of samples read */
{
  long i;
  int r,avail;
  unsigned char a,b;
  short s;

  if (AUdebugLevel > 1)
    printf("AuReadData: numSample=%ld\n",numSample);

  if (stream->write)
    CommonExit(1,"AuReadData: stream not in read mode");

  if (stream->fd==-1) {
    i = 0;
    while (i<numSample) {
      s = getshortB(stream);
      if (stream->eof)
	break;
      data[i++] = s;
    }
    stream->currentSample += i;
  }
  else {
    /* non-blocking */
    i = 0;
    avail = 1;
    while (i<numSample) {
      s = 0;
      if (avail) {
	r = read(stream->fd,&a,1);
	if (r==-1 && errno==EAGAIN)
	  avail = 0;
	else {
	  if (r!=1)
	    stream->eof = 1;
	  else {
	    do {
	      r = read(stream->fd,&b,1);
	    } while (r==-1 && errno==EAGAIN);   /* wait for 2nd byte */
	    if (r!=1)
	      stream->eof = 1;
	    else
	      s = (a<<8)|b;
	  }
	}
      }
      if (stream->eof)
	break;
      data[i++] = s;
    }
    stream->currentSample += i;
  }

  return i;
}


/* AuWriteData() */
/* Write data to audio stream. */
/* if non-blocking: discard samples that could not be written */

void AuWriteData (
  AuStream *stream,		/* in: audio stream (handle) */
  short *data,			/* in: data[] */
  long numSample)		/* in: number of samples to be written */
{
  long i;
  int r,avail;
  unsigned char a,b;
  short s;

  if (AUdebugLevel > 1)
    printf("AuWriteData: numSample=%ld\n",numSample);

  if (!stream->write)
    CommonExit(1,"AuWriteData: audio file not in write mode");

  if (stream->fd==-1) {
    for (i=0; i<numSample; i++)
      putshortB(data[i],stream);
    stream->currentSample += numSample;
  }
  else {
    /* non-blocking */
    avail = 1;
    for (i=0; i<numSample; i++) {
      s = data[i];
      a = (s>>8)&255;
      b = s&255;
      if (avail) {
	r = write(stream->fd,&a,1);
	if (r==-1 && errno==EAGAIN)
	  avail = 0;
	else {
	  if (r!=1)
	    stream->eof = 1;
	  else {
	    do {
	      r = write(stream->fd,&b,1);
	    } while (r==-1 && errno==EAGAIN);   /* wait for 2nd byte */
	    if (r!=1)
	      stream->eof = 1;
	  }
	}
      }
    }
    stream->currentSample += numSample;
  }

  if (stream->eof)
    CommonWarning("AuWriteDate: Can not write to au stream");
}


/* AuClose() */
/* Close audio stream.*/

void AuClose (
  AuStream *stream)		/* in: audio stream (handle) */
{
  if (AUdebugLevel)
    printf("AuClose: currentSample=%ld\n",stream->currentSample);

  if (stream->fd==-1) {
    if (stream->f!=stdin && stream->f!=stdout)
      fclose(stream->f);
  }
  else {
    if (stream->fd!=0 && stream->fd!=1)
      close(stream->fd);
  }
  free(stream);
}


/* end of austream.c */
